@@ -15,7 +15,10 @@
 # License along with this program; if not, write to the Free Software
 # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
+# pylint: disable=line-too-long
 """
+*New in pywbem 0.11 as experimental and finalized in 0.12.*
+
 Pywbem supports measuring the elapsed times of the WBEM operations that
 were performed in context of a connection, and maintaining a statistics
 over these times.
@@ -24,7 +27,7 @@ This capability is disabled by default and can be enabled in either
 of these ways:
 
 * When creating a :class:`~pywbem.WBEMConnection` object, via its
-  ``enable_stats`` argument.
+  ``stats_enabled`` argument.
 
 * After the :class:`~pywbem.WBEMConnection` object has been created, by
   modifying its :attr:`~pywbem.WBEMConnection.stats_enabled` instance
@@ -44,13 +47,102 @@ section for the operation names).
 The :class:`~pywbem.OperationStatistic` objects are under control of the
 :class:`~pywbem.Statistics` class.
 
-**Experimental:** The statistics support is experimental for this release.
-"""
+The statistics support maintains two kinds of times for each kind of WBEM
+operation:
+
+* Client times: The elapsed times for the WBEMConnection operation methods from
+  call to return. This is measured in pywbem pretty close to the API the pywbem
+  user is calling.
+
+* Server times: The elapsed times for processing the WBEM operations in the
+  WBEM server from receiving the CIM-XML request message to sending back the
+  CIM-XML response message. The server times are not measured by pywbem, but
+  are taken from the `WBEMServerResponseTime` HTTP header field of a CIM-XML
+  response, if present. See :term:`DSP0200` for a description of this header
+  field.
+
+  The `WBEMServerResponseTime` HTTP header field is optionally implemented
+  by WBEM servers. The following WBEM servers are known to implement this
+  header field:
+
+  * OpenPegasus
+
+The difference between client time and server time is the time spent in the
+pywbem client, plus the time spent on the network between client and server.
+
+The statistics support also maintains the size of the HTTP body in the CIM-XML
+request and response messages, in Bytes.
+
+These times and sizes are maintained as average, minimum and maximum values for
+each kind of operation in a connection.
+
+Finally, the statistics support maintains the total count of operations and the
+count of operations that failed, for each kind of operation.
+
+All data in the statistics applies to WBEM operations performed during periods
+of time where the statistics are enabled on a connection. Operations performed
+during periods of time where the statistics are disabled on a connection, are
+simply ignored in the statistics.
+
+For the Iter methods of WBEMConnection (e.g.
+:meth:`~pywbem.WbemConnection.IterEnumerateInstances`), the WBEM operations
+performed on behalf of them are subject of the statistics, but the Iter methods
+themselves do not show up in the statistics.
+
+The following example shows how statistics are enabled, and how statistics
+values are accessed individually using the
+:meth:`~pywbem.Statistics.get_op_statistic` method::
+
+    conn = pywbem.WBEMConnection(..., stats_enabled=True)
+
+    # Perform some operations on this connection
+    insts_1 = conn.EnumerateInstances('CIM_Foo1', 'root/cimv2')
+    insts_2 = conn.EnumerateInstances('CIM_Foo2', 'root/cimv2')
+    insts_3 = conn.EnumerateInstances('CIM_Foo3', 'root/cimv2')
+    inst_paths_1 = conn.EnumerateInstanceNames('CIM_Foo1', 'root/cimv2')
+
+    # Access statistics values for EnumerateInstances
+    ei_stats = conn.statistics.get_op_statistic('EnumerateInstances')
+    ei_count = ei_stats.count
+    ei_avg_client_time = ei_stats.avg_time
+    ei_avg_server_time = ei_stats.avg_server_time
+
+In the previous example, the values in ``ei_stats`` are "live", i.e. they
+continue to be updated as operations are performed. If a snapshot is needed at
+a certain point in time that remains unaffected by further operations, this can
+be achieved using the :meth:`~pywbem.Statistics.snapshot` method::
+
+    # Take snapshot and access statistics values for EnumerateInstances
+    stats_snapshot = dict(conn.statistics.snapshot())
+    ei_stats = stats_snapshot['EnumerateInstances']
+    ei_count = ei_stats.count
+    ei_avg_client_time = ei_stats.avg_time
+    ei_avg_server_time = ei_stats.avg_server_time
+
+It is also possible to simply print the current statistics of a connection as a
+formatted table, using the :meth:`~pywbem.Statistics.formatted` method::
+
+    # Print statistics values for all operations
+    print(conn.statistics.formatted())
+
+The output could look like this, if the WBEM server returns WBEM server
+response times::
+
+    Statistics (times in seconds, lengths in Bytes):
+    Count Excep         ClientTime              ServerTime             RequestLen                ReplyLen       Operation
+            Cnt     Avg     Min     Max     Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
+        3     0   0.234   0.100   0.401   0.204   0.080   0.361   1233   1000   1500    26667    20000    35000 EnumerateInstances
+        1     0   0.100   0.100   0.100   0.080   0.080   0.080   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
+"""  # noqa: E501
+# pylint: enable=line-too-long
 
 from __future__ import absolute_import
 
 import time
 import copy
+import six
+
+from ._utils import _format
 
 __all__ = ['Statistics', 'OperationStatistic']
 
@@ -58,21 +150,15 @@ __all__ = ['Statistics', 'OperationStatistic']
 class OperationStatistic(object):
     # pylint: disable=too-many-instance-attributes
     """
+    *New in pywbem 0.11 as experimental and finalized in 0.12.*
+
     A statistics data keeper for the executions of all operations with the
     same operation name.
 
-    This class maintains max, min, avg and count for elapsed time, request
-    size, and response size over the executed operations of a single
-    operation name.
-
-    Use the :meth:`pywbem.Statistics.start_timer` method to create objects
-    of this class::
-
-        stats = container.start_timer('EnumerateInstances')
-        ...
-        stats.stop_timer(request_len, reply_len, server_time, exc)
-
-    **Experimental:** This class is experimental for this release.
+    Objects of this class are created by the :class:`~pywbem.Statistics` class
+    and can be accessed by pywbem users through its
+    :meth:`~pywbem.Statistics.get_op_statistic` and
+    :meth:`~pywbem.Statistics.snapshot` methods.
     """
 
     def __init__(self, container, name):
@@ -124,9 +210,6 @@ class OperationStatistic(object):
         """
         :term:`string`: Name of the operation for which this statistics
         object maintains data.
-
-        This name is used by the :class:`~pywbem.Statistics` object
-        holding this time statistics as a key.
         """
         return self._name
 
@@ -141,25 +224,26 @@ class OperationStatistic(object):
     @property
     def count(self):
         """
-        :term:`integer`: The number of measured operations (that is,
-        invocations of the :meth:`~pywbem.OperationStatistic.stop_timer`
-        method).
+        :term:`integer`: The number of measured operations.
         """
         return self._count
 
     @property
     def exception_count(self):
         """
-        :term:`integer`: The number of exceptions that occurred when invoking
-        the measured operations.
+        :term:`integer`: The number of measured operations that resulted in an
+        exception returned to the pywbem user (for example because a failure
+        was indicated in the operation response of the WBEM server, or because
+        pywbem itself detected a failure before sending the request or after
+        receiving the response).
         """
         return self._exception_count
 
     @property
     def avg_time(self):
         """
-        :class:`py:float`: The average elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The average elapsed client time for execution of the
+        measured operations, in seconds.
         """
         try:
             return self._time_sum / self._count
@@ -169,24 +253,27 @@ class OperationStatistic(object):
     @property
     def min_time(self):
         """
-        :class:`py:float`: The minimum elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The minimum elapsed client time for execution of the
+        measured operations, in seconds.
         """
         return self._time_min
 
     @property
     def max_time(self):
         """
-        :class:`py:float`: The maximum elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The maximum elapsed client time for execution of the
+        measured operations, in seconds.
         """
         return self._time_max
 
     @property
     def avg_server_time(self):
         """
-        :class:`py:float`: The average elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The average elapsed server time for execution of the
+        measured operations, in seconds.
+
+        This time is 0 if the WBEM server did not return the WBEM server
+        response time.
         """
         try:
             return self._server_time_sum / self._count
@@ -196,16 +283,22 @@ class OperationStatistic(object):
     @property
     def min_server_time(self):
         """
-        :class:`py:float`: The minimum elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The minimum elapsed server time for execution of the
+        measured operations, in seconds.
+
+        This time is 0 if the WBEM server did not return the WBEM server
+        response time.
         """
         return self._server_time_min
 
     @property
     def max_server_time(self):
         """
-        :class:`py:float`: The maximum elapsed time for invoking the measured
-        operations, in seconds.
+        :class:`py:float`: The maximum elapsed server time for execution of the
+        measured operations, in seconds.
+
+        This time is 0 if the WBEM server did not return the WBEM server
+        response time.
         """
         return self._server_time_max
 
@@ -213,7 +306,7 @@ class OperationStatistic(object):
     def avg_request_len(self):
         """
         :class:`py:float`: The average size of the HTTP body in the CIM-XML
-        requests of the invoked operations, in Bytes.
+        requests of the measured operations, in Bytes.
         """
         try:
             return self._request_len_sum / self._count
@@ -224,7 +317,7 @@ class OperationStatistic(object):
     def min_request_len(self):
         """
         :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
-        requests of the invoked operations, in Bytes.
+        requests of the measured operations, in Bytes.
         """
         return self._request_len_min
 
@@ -232,7 +325,7 @@ class OperationStatistic(object):
     def max_request_len(self):
         """
         :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
-        requests of the invoked operations, in Bytes.
+        requests of the measured operations, in Bytes.
         """
         return self._request_len_max
 
@@ -240,7 +333,7 @@ class OperationStatistic(object):
     def avg_reply_len(self):
         """
         :class:`py:float`: The average size of the HTTP body in the CIM-XML
-        responses of the invoked operations, in Bytes.
+        responses of the measured operations, in Bytes.
         """
         try:
             return self._reply_len_sum / self._count
@@ -251,7 +344,7 @@ class OperationStatistic(object):
     def min_reply_len(self):
         """
         :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
-        responses of the invoked operations, in Bytes.
+        responses of the measured operations, in Bytes.
         """
         return self._reply_len_min
 
@@ -259,7 +352,7 @@ class OperationStatistic(object):
     def max_reply_len(self):
         """
         :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
-        responses of the invoked operations, in Bytes.
+        responses of the measured operations, in Bytes.
         """
         return self._reply_len_max
 
@@ -289,11 +382,9 @@ class OperationStatistic(object):
 
     def start_timer(self):
         """
-        This method needs to be called at the begin of an operation that
-        is intended to be measured. It starts the measurement for that
-        operation (by capturing the current point in time), if the
-        statistics container holding this object is enabled. Otherwise,
-        this method does nothing.
+        This is a low-level method that is called by pywbem at the begin of an
+        operation. It starts the measurement for that operation, if statistics
+        is enabled for the connection.
 
         A subsequent invocation of :meth:`~pywbem.OperationStatistic.stop_timer`
         will complete the measurement for that operation and will update the
@@ -307,29 +398,28 @@ class OperationStatistic(object):
     def stop_timer(self, request_len, reply_len, server_time=None,
                    exception=False):
         """
-        This method needs to be called at the end of an operation that is
-        intended to be measured. It completes the measurement for that
-        operation by capturing the needed data, and updates the statistics
-        data, if the statistics container holding this object is enabled.
-        Otherwise, this method does nothing.
+        This is a low-level method is called by pywbem at the end of an
+        operation. It completes the measurement for that operation by capturing
+        the needed data, and updates the statistics data, if statistics is
+        enabled for the connection.
 
         Parameters:
 
-          request_len (:term:`integer`)
+          request_len (:term:`integer`):
             Size of the HTTP body of the CIM-XML request message, in Bytes.
 
-          reply_len (:term:`integer`)
+          reply_len (:term:`integer`):
             Size of the HTTP body of the CIM-XML response message, in Bytes.
 
-          exception (:class:`py:bool`)
+          exception (:class:`py:bool`):
             Boolean that specifies whether an exception was raised while
             processing the operation.
 
-          server_time (:class:`py:bool`)
+          server_time (:class:`py:bool`):
             Time in seconds that the server optionally returns to the
             client in the HTTP response defining the time from when the
             server received the request to when it started sending the
-            response. If None, there is no time from the server.
+            response. If `None`, there is no time from the server.
 
         Returns:
 
@@ -337,97 +427,94 @@ class OperationStatistic(object):
           `None` if the statistics container holding this object is not
           enabled.
         """
-        if self.container.enabled:
-            if self._start_time is None:
-                raise RuntimeError('stop_timer() called without preceding '
-                                   ' start_timer()')
-            dt = time.time() - self._start_time
-            self._start_time = None
-            self._count += 1
-            self._time_sum += dt
-            self._request_len_sum += request_len
-            self._reply_len_sum += reply_len
-
-            if exception:
-                self._exception_count += 1
-
-            if dt > self._time_max:
-                self._time_max = dt
-            if dt < self._time_min:
-                self._time_min = dt
-
-            if server_time:
-                self._server_time_stored = True
-                self._server_time_sum += server_time
-                if dt > self._server_time_max:
-                    self._server_time_max = server_time
-                if dt < self._server_time_min:
-                    self._server_time_min = server_time
-
-            if request_len > self._request_len_max:
-                self._request_len_max = request_len
-            if request_len < self._request_len_min:
-                self._request_len_min = request_len
-
-            if reply_len > self._reply_len_max:
-                self._reply_len_max = reply_len
-            if reply_len < self._reply_len_min:
-                self._reply_len_min = reply_len
-
-            return dt
-        else:
+        if not self.container.enabled:
             return None
 
+        # stop the timer
+        if self._start_time is None:
+            raise RuntimeError('stop_timer() called without preceding '
+                               'start_timer()')
+        dt = time.time() - self._start_time
+        self._start_time = None
+        self._count += 1
+        self._time_sum += dt
+        self._request_len_sum += request_len
+        self._reply_len_sum += reply_len
+
+        if exception:
+            self._exception_count += 1
+
+        if dt > self._time_max:
+            self._time_max = dt
+        if dt < self._time_min:
+            self._time_min = dt
+
+        if server_time:
+            self._server_time_stored = True
+            self._server_time_sum += server_time
+            if dt > self._server_time_max:
+                self._server_time_max = server_time
+            if dt < self._server_time_min:
+                self._server_time_min = server_time
+
+        if request_len > self._request_len_max:
+            self._request_len_max = request_len
+        if request_len < self._request_len_min:
+            self._request_len_min = request_len
+
+        if reply_len > self._reply_len_max:
+            self._reply_len_max = reply_len
+        if reply_len < self._reply_len_min:
+            self._reply_len_min = reply_len
+
+        return dt
+
     def __repr__(self):
         """
         Return a human readable string with the statistics values, for debug
         purposes.
         """
-        return 'OperationStatistic(' \
-               'name={s.name!r}, ' \
-               'count={s.count!r}, ' \
-               'exception_count={s.exception_count!r}, ' \
-               'avg_time={s.avg_time!r}, ' \
-               'min_time={s.min_time!r}, ' \
-               'max_time={s.max_time!r}, ' \
-               'avg_server_time={s.avg_server_time!r}, ' \
-               'min_server_time={s.min_server_time!r}, ' \
-               'max_server_time={s.max_server_time!r}, ' \
-               'avg_request_len={s.avg_request_len!r}, ' \
-               'min_request_len={s.min_request_len!r}, ' \
-               'max_request_len={s.max_request_len!r}, ' \
-               'avg_reply_len={s.avg_reply_len!r}, ' \
-               'min_reply_len={s.min_reply_len!r}, ' \
-               'max_reply_len={s.max_reply_len!r})'. \
-               format(s=self)
+        return _format(
+            "OperationStatistic("
+            "name={s.name!A}, "
+            "count={s.count!A}, "
+            "exception_count={s.exception_count!A}, "
+            "avg_time={s.avg_time!A}, "
+            "min_time={s.min_time!A}, "
+            "max_time={s.max_time!A}, "
+            "avg_server_time={s.avg_server_time!A}, "
+            "min_server_time={s.min_server_time!A}, "
+            "max_server_time={s.max_server_time!A}, "
+            "avg_request_len={s.avg_request_len!A}, "
+            "min_request_len={s.min_request_len!A}, "
+            "max_request_len={s.max_request_len!A}, "
+            "avg_reply_len={s.avg_reply_len!A}, "
+            "min_reply_len={s.min_reply_len!A}, "
+            "max_reply_len={s.max_reply_len!A})",
+            s=self)
 
-    #: Formatted header string (two lines), for use with the formatted rows
-    #: returned by the :meth:`~pywbem.OperationStatistic.formatted` method.
-    #:
-    #: For an example, see :meth:`pywbem.Statistics.formatted`.
-    formatted_header_w_svr = \
-        'Count Excep            Time               ServerTime       ' \
-        '       RequestLen               ReplyLen         Operation\n' \
+    _formatted_header_w_svr = \
+        'Count Excep         ClientTime              ServerTime        ' \
+        '     RequestLen                ReplyLen       Operation\n' \
         '        Cnt     Avg     Min     Max     Avg     Min     Max   ' \
         ' Avg    Min    Max      Avg      Min      Max\n'
 
-    formatted_header = \
-        'Count Excep            Time          ' \
-        '       RequestLen            ReplyLen        Operation\n' \
+    _formatted_header = \
+        'Count Excep         ClientTime        ' \
+        '     RequestLen              ReplyLen       Operation\n' \
         '        Cnt     Avg     Min     Max   ' \
         ' Avg    Min    Max    Avg      Min      Max\n'
 
     def formatted(self, include_server_time):
         """
-        Return a formatted one-line string with the statistics values for this
-        operation.
+        Return a formatted one-line string with the statistics
+        values for the operation for which this statistics object
+        maintains data.
 
-        The returned string can be used with the formatted header string
-        defined in :attr:`~pywbem.OperationStatistic.formatted_header`.
-
-        For an example, see :meth:`pywbem.Statistics.formatted`.
+        This is a low-level method that is called by
+        :meth:`pywbem.Statistics.formatted`.
         """
-        if include_server_time:
+        if include_server_time:  # pylint: disable=no-else-return
             return ('{0:5d} {1:5d} '
                     '{2:7.3f} {3:7.3f} {4:7.3f} '
                     '{5:7.3f} {6:7.3f} {7:7.3f} '
@@ -458,25 +545,26 @@ class OperationStatistic(object):
 
 class Statistics(object):
     """
-    A container class for multiple operation statistic objects (of class
-    :class:`~pywbem.OperationStatistic`).
+    *New in pywbem 0.11 as experimental and finalized in 0.12.*
 
-    Each operation statistic object is identified by a name, that is defined
-    in the :meth:`~pywbem.Statistics.start_timer` method.
+    The statistics of a connection, that captures and provides statistics data
+    about the WBEM operations performed on the connection.
 
-    The statistics container can be in a state of enabled or disabled.
+    This class contains an operation statistic object (of class
+    :class:`~pywbem.OperationStatistic`) for each kind of WBEM operation.
+
+    A :class:`~pywbem.Statistics` object can be in a state of enabled or
+    disabled.
     If enabled, it accumulates the elapsed times between subsequent calls to
     the :meth:`~pywbem.OperationStatistic.start_timer` and
     :meth:`~pywbem.OperationStatistic.stop_timer` methods of class
     :class:`~pywbem.OperationStatistic`.
     If disabled, calls to these methods do not accumulate any time.
-    Initially, the statistics container is disabled. Its enablement state can
-    be controlled via the :meth:`~pywbem.Statistics.enable` and
-    :meth:`~pywbem.Statistics.disable` methods. Its current enablement
-    state can be accessed via the :attr:`~pywbem.Statistics.enabled`
-    property.
+    Initially, the statistics container is disabled.
 
-    **Experimental:** This class is experimental for this release.
+    The enablement state of the :class:`~pywbem.Statistics` object is
+    controlled by the statistics enablement state of the connection it belongs
+    to (see :meth:`pywbem.WBEMConnection.stats_enabled`)
     """
 
     def __init__(self, enable=False):
@@ -512,9 +600,14 @@ class Statistics(object):
 
     def start_timer(self, name):
         """
-        Start the timer for a given operation name and return the corresponding
+        This method is called by pywbem to start the timer for a particular
+        operation execution. It returns the corresponding
         :class:`~pywbem.OperationStatistic` object, creating one if needed.
 
+        The timer is subsequently stopped by pywbem by calling the
+        :meth:`~pywbem.OperationStatistic.stop_timer` method on the returned
+        :class:`~pywbem.OperationStatistic` object.
+
         Parameters:
 
           name (string):
@@ -576,7 +669,7 @@ class Statistics(object):
         """
         ret = "Statistics(\n"
         for name in self._op_stats:
-            ret += "  %r\n" % self._op_stats[name]
+            ret += _format("  {0!A}\n", self._op_stats[name])
         ret += ")"
         return ret
 
@@ -586,16 +679,22 @@ class Statistics(object):
         Return a human readable string with the statistics for this container.
         The operations are sorted by decreasing average time.
 
-        Server time statistic is included only if the wbem server has returned
-        statistics information.
+        The three columns for `ServerTime` are included only if the WBEM server
+        has returned WBEM server response times.
 
-        Example w/o server times::
+        Example if statistics are enabled::
 
             Statistics (times in seconds, lengths in Bytes):
-            Count  Exc            Time                ReqLen                ReplyLen          Operation
-                   Cnt    Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
-                3    0   0.234   0.100   0.401   1233   1000   1500    26667    20000    35000 EnumerateInstances
-                1    0   0.100   0.100   0.100   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
+            Count Excep         ClientTime              ServerTime             RequestLen                ReplyLen       Operation
+                    Cnt     Avg     Min     Max     Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
+                3     0   0.234   0.100   0.401   0.204   0.080   0.361   1233   1000   1500    26667    20000    35000 EnumerateInstances
+                1     0   0.100   0.100   0.100   0.080   0.080   0.080   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
+                . . .
+
+        Example if statistics are disabled::
+
+            Statistics (times in seconds, lengths in Bytes):
+            Disabled
         """  # noqa: E501
         # pylint: enable=line-too-long
         ret = "Statistics (times in seconds, lengths in Bytes):\n"
@@ -607,15 +706,33 @@ class Statistics(object):
             # Test to see if any server time is non-zero
             include_svr = False
             for name, stats in snapshot:  # pylint: disable=unused-variable
+                # pylint: disable=protected-access
                 if stats._server_time_stored:
                     include_svr = True
+            # pylint: disable=protected-access
             if include_svr:
-                ret += OperationStatistic.formatted_header_w_svr
+                ret += OperationStatistic._formatted_header_w_svr
             else:
-                ret += OperationStatistic.formatted_header
+                ret += OperationStatistic._formatted_header
 
             for name, stats in snapshot:  # pylint: disable=unused-variable
                 ret += stats.formatted(include_svr)
         else:
             ret += "Disabled"
         return ret.strip()
+
+    def reset(self):
+        """
+        Reset all statistics and clear any statistic names.
+        All statistics must be inactive before a reset will execute
+
+        Returns: True if reset, False if not
+        """
+        # Test for any stats being currently timed.
+        for stat in six.itervalues(self._op_stats):
+            if stat._start_time is not None:  # pylint: disable=protected-access
+                return False
+
+        # clear all statistics
+        self._op_stats = {}
+        return True
